@@ -30,21 +30,21 @@ from __future__ import absolute_import
 from ..util import TRANSPORT, parse_private_key, parse_certificate
 from ..piv import (
     PivController, ALGO, OBJ, SW, SLOT, PIN_POLICY, TOUCH_POLICY,
-    DEFAULT_MANAGEMENT_KEY)
+    DEFAULT_MANAGEMENT_KEY, generate_random_management_key)
 from ..driver_ccid import APDUError, SW_APPLICATION_NOT_FOUND
 from .util import (
     click_force_option, click_skip_on_help, click_callback, prompt_for_touch)
 from cryptography import x509
 from cryptography.exceptions import InvalidSignature
 from cryptography.hazmat.primitives import hashes, serialization
-from cryptography.hazmat.primitives.asymmetric import rsa, ec, padding
 from cryptography.hazmat.backends import default_backend
-from cryptography.x509.oid import NameOID
-from cryptography import utils
 from binascii import b2a_hex, a2b_hex
 import click
-import os
 import datetime
+import logging
+
+
+logger = logging.getLogger(__name__)
 
 
 @click_callback()
@@ -68,7 +68,11 @@ def click_parse_format(ctx, param, val):
 @click_callback()
 def click_parse_management_key(ctx, param, val):
     try:
-        return a2b_hex(val)
+        key = a2b_hex(val)
+        if key and len(key) != 24:
+            return ValueError('Management key must be exactly 24 bytes '
+                              '(48 hexadecimal digits) long.')
+        return key
     except Exception:
         return ValueError(val)
 
@@ -98,14 +102,14 @@ click_touch_policy_option = click.option(
 @click_skip_on_help
 def piv(ctx):
     """
-    Manage YubiKey PIV functionality.
+    Manage PIV application.
     """
     try:
         controller = PivController(ctx.obj['dev'].driver)
         ctx.obj['controller'] = controller
     except APDUError as e:
         if e.sw == SW_APPLICATION_NOT_FOUND:
-            ctx.fail("The PIV functionality can't be found on this YubiKey.")
+            ctx.fail("The PIV application can't be found on this YubiKey.")
         raise
 
 
@@ -113,7 +117,7 @@ def piv(ctx):
 @click.pass_context
 def info(ctx):
     """
-    Display status of PIV functionality.
+    Display status of PIV application.
     """
     controller = ctx.obj['controller']
     click.echo('PIV version: %d.%d.%d' % controller.version)
@@ -167,8 +171,8 @@ def reset(ctx):
     """
     Reset all PIV data.
 
-    This action will wipe all data and reset factory settings for
-    the PIV functionality on your YubiKey.
+    This action will wipe all data and restore factory settings for
+    the PIV application on your YubiKey.
     """
 
     click.echo('Resetting PIV data...')
@@ -361,7 +365,8 @@ def attest(ctx, slot, certificate, format):
     controller = ctx.obj['controller']
     try:
         cert = controller.attest(slot)
-    except APDUError:
+    except APDUError as e:
+        logger.error('Attestation failed', exc_info=e)
         ctx.fail('Attestation failed.')
     certificate.write(cert.public_bytes(encoding=format))
 
@@ -387,6 +392,9 @@ def export_certificate(ctx, slot, format, certificate):
     except APDUError as e:
         if e.sw == SW.NOT_FOUND:
             ctx.fail('No certificate found.')
+        else:
+            logger.error('Failed to read certificate from slot %s', slot,
+                         exc_info=e)
     certificate.write(cert.public_bytes(encoding=format))
 
 
@@ -432,7 +440,8 @@ def set_pin_retries(ctx, management_key, pin, pin_retries, puk_retries, force):
     """
     controller = ctx.obj['controller']
     _ensure_authenticated(
-        ctx, controller, pin, management_key, require_pin_and_key=True)
+        ctx, controller, pin, management_key, require_pin_and_key=True,
+        no_prompt=force)
     click.echo('WARNING: This will reset the PIN and PUK to the factory '
                'defaults!')
     force or click.confirm('Set PIN and PUK retry counters to: {} {}?'.format(
@@ -442,7 +451,8 @@ def set_pin_retries(ctx, management_key, pin, pin_retries, puk_retries, force):
         click.echo('Default PINs are set.')
         click.echo('PIN:    123456')
         click.echo('PUK:    12345678')
-    except Exception:
+    except Exception as e:
+        logger.error('Failed to set PIN retries', exc_info=e)
         ctx.fail('Setting pin retries failed.')
 
 
@@ -479,47 +489,21 @@ def generate_certificate(
     public_key = serialization.load_pem_public_key(
         data, default_backend())
 
-    algorithm = ALGO.from_public_key(public_key)
-
-    builder = x509.CertificateBuilder()
-    builder = builder.public_key(public_key)
-    builder = builder.subject_name(
-        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, subject), ]))
-
-    # Same as subject on self-signed certificates.
-    builder = builder.issuer_name(
-        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, subject), ]))
-
-    # x509.random_serial_number added in cryptography 1.6
-    serial = utils.int_from_bytes(os.urandom(20), 'big') >> 1
-    builder = builder.serial_number(serial)
-
     now = datetime.datetime.now()
-    builder = builder.not_valid_before(now)
-    builder = builder.not_valid_after(now + datetime.timedelta(days=valid_days))
+    valid_to = now + datetime.timedelta(days=valid_days)
 
     try:
-        cert = controller.sign_cert_builder(
-            slot, algorithm, builder, touch_callback=prompt_for_touch)
-    except APDUError:
+        controller.generate_self_signed_certificate(
+            slot, public_key, subject, now, valid_to,
+            touch_callback=prompt_for_touch)
+
+    except APDUError as e:
+        logger.error('Failed to generate certificate for slot %s', slot,
+                     exc_info=e)
         ctx.fail('Certificate generation failed.')
 
-    # Verify that the public key used in the certificate
-    # is from the same keypair as the private key.
-    cert_signature = cert.signature
-    cert_bytes = cert.tbs_certificate_bytes
-    if isinstance(public_key, rsa.RSAPublicKey):
-        verifier = public_key.verifier(
-            cert_signature, padding.PKCS1v15(), cert.signature_hash_algorithm)
-    elif isinstance(public_key, ec.EllipticCurvePublicKey):
-        verifier = public_key.verifier(
-            cert_signature, ec.ECDSA(cert.signature_hash_algorithm))
-    verifier.update(cert_bytes)
-    try:
-        verifier.verify()
     except InvalidSignature:
         ctx.fail('Invalid signature, certificate not imported.')
-    controller.import_certificate(slot, cert)
 
 
 @piv.command('generate-csr')
@@ -550,15 +534,12 @@ def generate_certificate_signing_request(
     public_key = serialization.load_pem_public_key(
         data, default_backend())
 
-    builder = x509.CertificateSigningRequestBuilder()
-    builder = builder.subject_name(
-        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, subject), ]))
-
     try:
-        csr = controller.sign_csr_builder(
-            slot, public_key, builder, touch_callback=prompt_for_touch)
+        csr = controller.generate_certificate_signing_request(
+            slot, public_key, subject, touch_callback=prompt_for_touch)
     except APDUError:
         ctx.fail('Certificate Signing Request generation failed.')
+
     csr_output.write(csr.public_bytes(encoding=serialization.Encoding.PEM))
 
 
@@ -600,7 +581,8 @@ def change_pin(ctx, pin, new_pin):
             show_default=False, confirmation_prompt=True)
     try:
         controller.change_pin(pin, new_pin)
-    except APDUError:
+    except APDUError as e:
+        logger.error('Failed to change PIN', exc_info=e)
         ctx.fail('Changing the PIN failed.')
     click.echo('New PIN set.')
 
@@ -622,11 +604,14 @@ def change_puk(ctx, puk, new_puk):
         new_puk = click.prompt(
             'Enter your new PUK', default='', hide_input=True,
             show_default=False, confirmation_prompt=True)
-    try:
-        controller.change_puk(puk, new_puk)
-    except APDUError:
-        ctx.fail('Changing the PUK failed.')
-    click.echo('New PUK set.')
+
+    (success, retries) = controller.change_puk(puk, new_puk)
+
+    if success:
+        click.echo('New PUK set.')
+    else:
+        logger.debug('Failed to change PUK, %d tries left', retries)
+        ctx.fail('PUK change failed - %d tries left.' % retries)
 
 
 @piv.command('change-management-key')
@@ -635,7 +620,9 @@ def change_puk(ctx, puk, new_puk):
 @click.option(
     '-t', '--touch', is_flag=True,
     help='Require touch on YubiKey when prompted for management key.')
-@click.option('-n', '--new-management-key', help='A new management key.')
+@click.option(
+    '-n', '--new-management-key', help='A new management key.',
+    callback=click_parse_management_key)
 @click.option(
     '-m', '--management-key', help='Current management key.',
     callback=click_parse_management_key)
@@ -643,9 +630,14 @@ def change_puk(ctx, puk, new_puk):
     '-p', '--protect', is_flag=True,
     help='Store new management key on your YubiKey, protected by PIN.'
          ' A random key will be used if no key is provided.')
+@click.option(
+    '-g', '--generate', is_flag=True, help='Generate a random management key. '
+    'Implied by --protect unless --new-management-key is also given. '
+    'Conflicts with --new-management-key.')
 @click_force_option
 def change_management_key(
-        ctx, management_key, pin, new_management_key, touch, protect, force):
+        ctx, management_key, pin, new_management_key, touch, protect, generate,
+        force):
     """
     Change the management key.
 
@@ -659,7 +651,12 @@ def change_management_key(
         ctx, controller, pin, management_key,
         require_pin_and_key=protect,
         mgm_key_prompt='Enter your current management key '
-                       '[blank to use default key]')
+                       '[blank to use default key]',
+        no_prompt=force)
+
+    if new_management_key and generate:
+        ctx.fail('Invalid options: --new-management-key conflicts with '
+                 '--generate')
 
     # Touch not supported on NEO.
     if touch and controller.version < (4, 0, 0):
@@ -668,39 +665,43 @@ def change_management_key(
     # If an old stored key needs to be cleared, the PIN is needed.
     if not protect and controller.has_stored_key:
         if pin:
-            _verify_pin(ctx, controller, pin)
+            _verify_pin(ctx, controller, pin, no_prompt=force)
         else:
             force or click.confirm(
                     'The current management key is stored on the YubiKey'
                     ' and will not be cleared if no PIN is provided. Continue?',
                     abort=True)
 
-    # If the key should be protected by PIN and no key is given,
-    # we generate a random key.
-    if protect and not new_management_key:
-        new_management_key = _generate_random_management_key()
+    if not new_management_key and not protect:
+        if generate:
+            new_management_key = generate_random_management_key()
 
-    if not new_management_key:
-        if not force:
+            if not protect:
+                click.echo(
+                    'Generated management key: {}'.format(
+                        b2a_hex(new_management_key).decode('utf-8')))
+
+        elif force:
+            ctx.fail('New management key not given. Please remove the --force '
+                     'flag, or set the --generate flag or the '
+                     '--new-management-key option.')
+
+        else:
             new_management_key = click.prompt(
-                'Enter your new management key'
-                ' [blank to randomly generate]',
-                default='', show_default=False,
+                'Enter your new management key',
                 hide_input=True, confirmation_prompt=True)
 
-        if force or new_management_key == '':
-            new_management_key = _generate_random_management_key()
-            click.echo(
-                'Generated management key: {}'.format(new_management_key))
+    if new_management_key and type(new_management_key) is not bytes:
+        try:
+            new_management_key = a2b_hex(new_management_key)
+        except Exception:
+            ctx.fail('New management key has the wrong format.')
 
-    try:
-        new_management_key = a2b_hex(new_management_key)
-    except Exception:
-        ctx.fail('New management key has the wrong format.')
     try:
         controller.set_mgm_key(
             new_management_key, touch=touch, store_on_device=protect)
-    except APDUError:
+    except APDUError as e:
+        logger.error('Failed to change management key', exc_info=e)
         ctx.fail('Changing the management key failed.')
 
 
@@ -736,10 +737,6 @@ def _prompt_management_key(
         ctx.fail('Management key has the wrong format.')
 
 
-def _generate_random_management_key():
-    return b2a_hex(os.urandom(24)).decode()
-
-
 def _prompt_pin(ctx, prompt='Enter PIN'):
     return click.prompt(
         prompt, default='', hide_input=True, show_default=False)
@@ -748,34 +745,45 @@ def _prompt_pin(ctx, prompt='Enter PIN'):
 def _ensure_authenticated(
         ctx, controller, pin=None, management_key=None,
         require_pin_and_key=False,
-        mgm_key_prompt=None):
+        mgm_key_prompt=None,
+        no_prompt=False):
 
     if controller.has_protected_key:
         if not management_key:
-            _verify_pin(ctx, controller, pin)
+            _verify_pin(ctx, controller, pin, no_prompt=no_prompt)
         else:
-            _authenticate(ctx, controller, management_key, mgm_key_prompt)
+            _authenticate(ctx, controller, management_key, mgm_key_prompt,
+                          no_prompt=no_prompt)
     else:
         if require_pin_and_key:
-            _verify_pin(ctx, controller, pin)
-        _authenticate(ctx, controller, management_key, mgm_key_prompt)
+            _verify_pin(ctx, controller, pin, no_prompt=no_prompt)
+        _authenticate(ctx, controller, management_key, mgm_key_prompt,
+                      no_prompt=no_prompt)
 
 
-def _verify_pin(ctx, controller, pin):
+def _verify_pin(ctx, controller, pin, no_prompt=False):
     if not pin:
-        pin = _prompt_pin(ctx)
+        if no_prompt:
+            ctx.fail('PIN required.')
+        else:
+            pin = _prompt_pin(ctx)
+
     try:
         controller.verify(pin, touch_callback=prompt_for_touch)
     except APDUError:
         ctx.fail('PIN verification failed.')
 
 
-def _authenticate(ctx, controller, management_key, mgm_key_prompt):
+def _authenticate(ctx, controller, management_key, mgm_key_prompt,
+                  no_prompt=False):
     if not management_key:
-        if mgm_key_prompt is None:
-            management_key = _prompt_management_key(ctx)
+        if no_prompt:
+            ctx.fail('Management key required.')
         else:
-            management_key = _prompt_management_key(ctx, mgm_key_prompt)
+            if mgm_key_prompt is None:
+                management_key = _prompt_management_key(ctx)
+            else:
+                management_key = _prompt_management_key(ctx, mgm_key_prompt)
     try:
         controller.authenticate(management_key, touch_callback=prompt_for_touch)
     except APDUError:
@@ -789,20 +797,17 @@ def _check_eccp384(ctx, controller, algorithm):
 
 
 def _check_pin_policy(ctx, controller, pin_policy):
-    #  Pin policy not supported on NEO.
-    if pin_policy is not None and controller.version < (4, 0, 0):
+    if pin_policy is not None and not controller.supports_pin_policies:
         ctx.fail('Pin policy is not supported by this YubiKey.')
 
 
 def _check_touch_policy(ctx, controller, touch_policy):
-    #  Touch policy not supported on NEO.
     if touch_policy is not None:
-        if controller.version < (4, 0, 0):
+        if len(controller.supported_touch_policies) == 0:
             ctx.fail('Touch policy is not supported by this YubiKey.')
-        if touch_policy == TOUCH_POLICY.CACHED \
-                and controller.version < (4, 3, 0):
-            #  Cached policy was added in 4.3
-            ctx.fail('Touch policy "CACHED" not supported by this YubiKey.')
+        elif touch_policy not in controller.supported_touch_policies:
+            ctx.fail('Touch policy {} not supported by this YubiKey.'.format(
+                touch_policy.name))
 
 
 piv.transports = TRANSPORT.CCID
