@@ -27,16 +27,18 @@
 
 from __future__ import absolute_import
 
-import six
 import time
+import logging
 from .native.ykpers import Ykpers
 from ctypes import sizeof, byref, c_int, c_uint, c_size_t, create_string_buffer
 from .driver import AbstractDriver, ModeSwitchError
 from .util import (PID, TRANSPORT, MissingLibrary, time_challenge,
                    parse_totp_hash, format_code, hmac_shorten_key)
-from .scanmap import us
+from .scancodes import encode, KEYBOARD_LAYOUT
 from binascii import a2b_hex, b2a_hex
 
+logger = logging.getLogger(__name__)
+
 INS_SELECT = 0xa4
 INS_YK4_CAPABILITIES = 0x1d
 
@@ -56,7 +58,8 @@ try:
         raise Exception('yk_init failed.')
     libversion = tuple(int(x) for x in ykpers.ykpers_check_version(None)
                        .decode('ascii').split('.'))
-except Exception:
+except Exception as e:
+    logger.error('libykpers not found', exc_info=e)
     ykpers = MissingLibrary(
         'libykpers not found, slot functionality not available!')
     libversion = None
@@ -87,12 +90,6 @@ def slot_to_cmd(slot, update=False):
         raise ValueError('slot must be 1 or 2')
 
 
-def get_scan_codes(ascii):
-    if isinstance(ascii, six.text_type):
-        ascii = ascii.encode('ascii')
-    return bytes(bytearray(us.scancodes[c] for c in six.iterbytes(ascii)))
-
-
 class OTPDriver(AbstractDriver):
     """
     libykpers based OTP driver
@@ -128,7 +125,9 @@ class OTPDriver(AbstractDriver):
         if ykpers.yk_get_serial(self._dev, 0, 0, byref(serial)):
             return serial.value
         else:
-            return None  # Serial not visible
+            logger.debug(
+                'Failed to read serial from device.')
+        return None  # Serial not visible
 
     def _read_status(self):
         status = ykpers.ykds_alloc()
@@ -213,7 +212,9 @@ class OTPDriver(AbstractDriver):
         finally:
             ykpers.ykp_free_config(cfg)
 
-    def program_static(self, slot, password, append_cr=True):
+    def program_static(
+        self, slot, password, append_cr=True,
+            keyboard_layout=KEYBOARD_LAYOUT.MODHEX):
         pw_len = len(password)
         if self._version < (2, 0, 0):
             raise ValueError('static password requires YubiKey 2.0.0 or later')
@@ -233,7 +234,7 @@ class OTPDriver(AbstractDriver):
             if append_cr:
                 check(ykpers.ykp_set_tktflag(cfg, 'APPEND_CR'))
 
-            pw_bytes = get_scan_codes(password)
+            pw_bytes = encode(password, keyboard_layout=keyboard_layout)
             if pw_len <= 16:  # All in fixed
                 check(ykpers.ykp_set_fixed(cfg, pw_bytes, pw_len))
             elif pw_len <= 16 + 6:  # All in fixed and uid
@@ -286,6 +287,10 @@ class OTPDriver(AbstractDriver):
         # Give the YubiKey 10 tries to do the calculation.
         for idx in range(10):
             try:
+                logger.debug(
+                    'Sending a challenge to the device. Slot %s. '
+                    'Attempt %s. Wait for touch is %s.', slot, idx + 1,
+                    wait_for_touch)
                 check(ykpers.yk_challenge_response(
                         self._dev, SLOTS[slot], wait_for_touch,
                         len(challenge), challenge, sizeof(resp), resp))
@@ -294,8 +299,17 @@ class OTPDriver(AbstractDriver):
                     # Error 11 when wait_for_touch is true is an unexpected
                     # state, let's try again.
                     continue
-                else:
+                elif wait_for_touch is False:
+                    logger.debug('Got %s as expected.', e)
+                    # NEOs and very old YK4s might still be blinking,
+                    # lets try to read the serial to cancel it.
+                    self._read_serial()
                     raise
+                else:
+                    logger.debug('YkpersError: %s', e)
+                    raise
+            # We got a result, break the loop.
+            break
         if totp:
             return format_code(parse_totp_hash(resp.raw[:20]), digits)
         else:
@@ -378,8 +392,15 @@ class OTPDriver(AbstractDriver):
         finally:
             ykpers.ykp_free_config(cfg)
 
+    def close(self):
+        if self._dev is not None:
+            logger.debug('Close %s', self)
+            ykpers.yk_close_key(self._dev)
+            self._dev = None
+
     def __del__(self):
-        ykpers.yk_close_key(self._dev)
+        logger.debug('Destroy %s', self)
+        self.close()
 
 
 def open_devices():
@@ -389,5 +410,7 @@ def open_devices():
         for i in range(255):
             dev = ykpers.yk_open_key(i)
             if not dev:
+                logger.debug('Failed to open key at position %s', i)
                 break
+            logger.debug('Success in opening key at position %s', i)
             yield OTPDriver(dev)
