@@ -32,22 +32,28 @@ from .driver_ccid import APDUError, SW_OK, SW_APPLICATION_NOT_FOUND
 from .util import (
     AID, Tlv, parse_tlvs,
     ensure_not_cve201715361_vulnerable_firmware_version)
+from collections import namedtuple
 from cryptography import x509
 from cryptography.utils import int_to_bytes, int_from_bytes
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 from cryptography.hazmat.primitives.constant_time import bytes_eq
-from cryptography.hazmat.primitives.asymmetric import rsa, ec
+from cryptography.hazmat.primitives.asymmetric import rsa, ec, padding
 from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
 from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
 from cryptography.hazmat.backends import default_backend
+from cryptography.x509.oid import NameOID
 from collections import OrderedDict
 from threading import Timer
+import logging
 import struct
 import six
 import os
 
 
+logger = logging.getLogger(__name__)
+
+
 @unique
 class INS(IntEnum):
     VERIFY = 0x20
@@ -248,6 +254,10 @@ class SW(IntEnum):
     NOT_FOUND = 0x6a82
     ACCESS_DENIED = 0x6982
     AUTHENTICATION_BLOCKED = 0x6983
+    INCORRECT_PARAMETERS = 0x6a80
+
+
+CodeChangeResult = namedtuple('CodeChangeResult', ['success', 'tries_left'])
 
 
 PIN = 0x80
@@ -350,6 +360,10 @@ def _derive_key(pin, salt):
     return kdf.derive(pin.encode('utf-8'))
 
 
+def generate_random_management_key():
+    return os.urandom(24)
+
+
 class PivmanData(object):
 
     def __init__(self, raw_data=Tlv(0x80)):
@@ -416,6 +430,9 @@ class PivController(object):
         self._authenticated = False
         self._driver = driver
         self._version = self._read_version()
+        self._update_pivman_data()
+
+    def _update_pivman_data(self):
         try:
             self._pivman_data = PivmanData(self.get_data(OBJ.PIVMAN_DATA))
         except APDUError:
@@ -484,10 +501,12 @@ class PivController(object):
         if self.has_derived_key and not self._authenticated:
             self.authenticate(
                 _derive_key(pin, self._pivman_data.salt), touch_callback)
+            self.verify(pin, touch_callback)
 
         if self.has_stored_key and not self._authenticated:
             self._init_pivman_protected()
             self.authenticate(self._pivman_protected_data.key, touch_callback)
+            self.verify(pin, touch_callback)
 
     def change_pin(self, old_pin, new_pin):
         self.send_cmd(INS.CHANGE_REFERENCE, 0, PIN,
@@ -498,8 +517,15 @@ class PivController(object):
             self.use_derived_key(new_pin)
 
     def change_puk(self, old_puk, new_puk):
-        self.send_cmd(INS.CHANGE_REFERENCE, 0, PUK,
-                      _pack_pin(old_puk) + _pack_pin(new_puk))
+        try:
+            self.send_cmd(INS.CHANGE_REFERENCE, 0, PUK,
+                          _pack_pin(old_puk) + _pack_pin(new_puk))
+            return CodeChangeResult(True, None)
+        except APDUError as e:
+            retries = self._parse_tries_left(e.sw)
+            logger.debug('PUK change failed, %d tries remaining', retries,
+                         exc_info=e)
+            return CodeChangeResult(False, retries)
 
     def unblock_pin(self, puk, new_pin):
         try:
@@ -561,6 +587,30 @@ class PivController(object):
         self._authenticated = True
 
     def set_mgm_key(self, new_key, touch=False, store_on_device=False):
+        # If the key should be protected by PIN and no key is given,
+        # we generate a random key.
+        if not new_key:
+            if store_on_device:
+                new_key = generate_random_management_key()
+            else:
+                raise ValueError('new_key was not given and '
+                                 'store_on_device was not True')
+
+        if len(new_key) != 24:
+            raise ValueError('Management key must be exactly 24 bytes long, '
+                             'was: {}'.format(len(new_key)))
+
+        if store_on_device or (not store_on_device and self.has_stored_key):
+            # Ensure we have access to protected data before overwriting key
+            try:
+                self._init_pivman_protected()
+            except Exception as e:
+                logger.debug('Failed to initialize protected pivman data',
+                             exc_info=e)
+
+                if store_on_device:
+                    raise
+
         # Set the new management key
         self.send_cmd(
             INS.SET_MGMKEY, 0xff, 0xfe if touch else 0xff,
@@ -574,7 +624,6 @@ class PivController(object):
         self.put_data(OBJ.PIVMAN_DATA, self._pivman_data.get_bytes())
         if store_on_device:
             # Store key in protected pivman data
-            self._init_pivman_protected()
             self._pivman_protected_data.key = new_key
             self.put_data(
                 OBJ.PIVMAN_PROTECTED_DATA,
@@ -583,14 +632,12 @@ class PivController(object):
             # If new key should not be stored and there is an old stored key,
             # try to clear it.
             try:
-                self._init_pivman_protected()
                 self._pivman_protected_data.key = None
                 self.put_data(
                     OBJ.PIVMAN_PROTECTED_DATA,
                     self._pivman_protected_data.get_bytes())
-            except APDUError:
-                # No pin provided, can't clear key..
-                pass
+            except APDUError as e:
+                logger.debug("No PIN provided, can't clear key..", exc_info=e)
 
     def get_pin_tries(self):
         """
@@ -598,19 +645,16 @@ class PivController(object):
         0 PIN authentication blocked. Note that 15 is the highest
         value that will be returned even if remaining tries is higher.
         """
-        try:
-            # Verify without PIN gives number of tries left.
-            _, sw = self.send_cmd(INS.VERIFY, 0, PIN)
-        except APDUError as e:
-            return self._parse_tries_left(e.sw)
+        # Verify without PIN gives number of tries left.
+        _, sw = self.send_cmd(INS.VERIFY, 0, PIN, check=None)
+        return self._parse_tries_left(sw)
 
     def _get_puk_tries(self):
-        try:
-            # A failed unblock pin will return number of PUK tries left,
-            # but also uses one try.
-            _, sw = self.send_cmd(INS.RESET_RETRY, 0, PIN, _pack_pin('')*2)
-        except APDUError as e:
-            return self._parse_tries_left(e.sw)
+        # A failed unblock pin will return number of PUK tries left,
+        # but also uses one try.
+        _, sw = self.send_cmd(INS.RESET_RETRY, 0, PIN, _pack_pin('')*2,
+                              check=None)
+        return self._parse_tries_left(sw)
 
     def _parse_tries_left(self, sw):
         # Blocked, 0 tries left.
@@ -636,6 +680,7 @@ class PivController(object):
         self._block_pin()
         self._block_puk()
         self.send_cmd(INS.RESET)
+        self._update_pivman_data()
 
     def get_data(self, object_id):
         id_bytes = struct.pack(b'>I', object_id).lstrip(b'\0')
@@ -677,6 +722,66 @@ class PivController(object):
             ).public_key(default_backend())
         raise ValueError('Invalid algorithm!')
 
+    def generate_self_signed_certificate(
+            self, slot, public_key, common_name, valid_from, valid_to,
+            touch_callback=None):
+
+        algorithm = ALGO.from_public_key(public_key)
+
+        builder = x509.CertificateBuilder()
+        builder = builder.public_key(public_key)
+        builder = builder.subject_name(
+            x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name), ]))
+
+        # Same as subject on self-signed certificates.
+        builder = builder.issuer_name(
+            x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name), ]))
+
+        # x509.random_serial_number added in cryptography 1.6
+        serial = int_from_bytes(os.urandom(20), 'big') >> 1
+        builder = builder.serial_number(serial)
+
+        builder = builder.not_valid_before(valid_from)
+        builder = builder.not_valid_after(valid_to)
+
+        try:
+            cert = self.sign_cert_builder(
+                slot, algorithm, builder, touch_callback)
+        except APDUError as e:
+            logger.error('Failed to generate certificate for slot %s', slot,
+                         exc_info=e)
+            raise
+
+        # Verify that the public key used in the certificate
+        # is from the same keypair as the private key.
+        cert_signature = cert.signature
+        cert_bytes = cert.tbs_certificate_bytes
+        if isinstance(public_key, rsa.RSAPublicKey):
+            verifier = public_key.verifier(
+                cert_signature, padding.PKCS1v15(),
+                cert.signature_hash_algorithm)
+        elif isinstance(public_key, ec.EllipticCurvePublicKey):
+            verifier = public_key.verifier(
+                cert_signature, ec.ECDSA(cert.signature_hash_algorithm))
+        verifier.update(cert_bytes)
+        verifier.verify()
+        self.import_certificate(slot, cert)
+
+    def generate_certificate_signing_request(self, slot, public_key, subject,
+                                             touch_callback=None):
+        builder = x509.CertificateSigningRequestBuilder()
+        builder = builder.subject_name(
+            x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, subject), ]))
+
+        try:
+            return self.sign_csr_builder(
+                slot, public_key, builder, touch_callback=touch_callback)
+        except APDUError as e:
+            logger.error(
+                'Failed to generate Certificate Signing Request for slot %s',
+                slot, exc_info=e)
+            raise
+
     def import_key(self, slot, key, pin_policy=PIN_POLICY.DEFAULT,
                    touch_policy=TOUCH_POLICY.DEFAULT):
         algorithm, data = _get_key_data(key)
@@ -719,12 +824,19 @@ class PivController(object):
         return self._raw_sign_decrypt(slot, algorithm, Tlv(0x81, message),
                                       _sign_len_conditions[algorithm])
 
+    def sign(self, slot, algorithm, message):
+        if algorithm in (ALGO.RSA1024, ALGO.RSA2048):
+            message = _pkcs1_15_pad(algorithm, message)
+        elif algorithm in (ALGO.ECCP256, ALGO.ECCP384):
+            h = hashes.Hash(hashes.SHA256(), default_backend())
+            h.update(message)
+            message = h.finalize()
+        return self.sign_raw(slot, algorithm, message)
+
     def decrypt_raw(self, slot, algorithm, message):
         return self._raw_sign_decrypt(slot, algorithm, Tlv(0x85, message),
                                       _decrypt_len_conditions[algorithm])
 
-    # Not sure if these should go in this class or somewhere else
-
     def list_certificates(self):
         certs = OrderedDict()
         for slot in set(SLOT) - {SLOT.CARD_MANAGEMENT, SLOT.ATTESTATION}:
@@ -766,20 +878,12 @@ class PivController(object):
     def sign_cert_builder(self, slot, algorithm, builder, touch_callback=None):
         dummy_key = _dummy_key(algorithm)
         cert = builder.sign(dummy_key, hashes.SHA256(), default_backend())
-        message = cert.tbs_certificate_bytes
-
-        if algorithm in (ALGO.RSA1024, ALGO.RSA2048):
-            message = _pkcs1_15_pad(algorithm, message)
-        elif algorithm in (ALGO.ECCP256, ALGO.ECCP384):
-            h = hashes.Hash(hashes.SHA256(), default_backend())
-            h.update(message)
-            message = h.finalize()
 
         if touch_callback is not None:
             touch_timer = Timer(0.500, touch_callback)
             touch_timer.start()
 
-        sig = self.sign_raw(slot, algorithm, message)
+        sig = self.sign(slot, algorithm, cert.tbs_certificate_bytes)
 
         if touch_callback is not None:
             touch_timer.cancel()
@@ -795,39 +899,43 @@ class PivController(object):
     def sign_csr_builder(self, slot, public_key, builder, touch_callback=None):
         algorithm = ALGO.from_public_key(public_key)
         dummy_key = _dummy_key(algorithm)
-        cert = builder.sign(dummy_key, hashes.SHA256(), default_backend())
-        message = cert.tbs_certrequest_bytes
+        csr = builder.sign(dummy_key, hashes.SHA256(), default_backend())
+        seq = parse_tlvs(Tlv(csr.public_bytes(Encoding.DER)).value)
 
-        if algorithm in (ALGO.RSA1024, ALGO.RSA2048):
-            message = _pkcs1_15_pad(algorithm, message)
-            dummy_bytes = dummy_key.public_key().public_bytes(
-                Encoding.DER, PublicFormat.PKCS1)
-            pub_bytes = public_key.public_bytes(
-                Encoding.DER, PublicFormat.PKCS1)
-        elif algorithm in (ALGO.ECCP256, ALGO.ECCP384):
-            h = hashes.Hash(hashes.SHA256(), default_backend())
-            h.update(message)
-            message = h.finalize()
-            dummy_bytes = dummy_key.public_key().public_bytes(
-                Encoding.DER, PublicFormat.SubjectPublicKeyInfo)
-            pub_bytes = public_key.public_bytes(
-                Encoding.DER, PublicFormat.SubjectPublicKeyInfo)
+        # Replace public key
+        pub_format = PublicFormat.PKCS1 if algorithm.name.startswith('RSA') \
+            else PublicFormat.SubjectPublicKeyInfo
+        dummy_bytes = dummy_key.public_key().public_bytes(
+            Encoding.DER, pub_format)
+        pub_bytes = public_key.public_bytes(Encoding.DER, pub_format)
+        seq[0] = seq[0].replace(dummy_bytes, pub_bytes)
 
         if touch_callback is not None:
             touch_timer = Timer(0.500, touch_callback)
             touch_timer.start()
 
-        sig = self.sign_raw(slot, algorithm, message)
+        sig = self.sign(slot, algorithm, seq[0])
 
         if touch_callback is not None:
             touch_timer.cancel()
 
-        seq = parse_tlvs(Tlv(cert.public_bytes(Encoding.DER)).value)
-        # Replace public key
-        seq[0] = seq[0].replace(dummy_bytes, pub_bytes)
         # Replace signature, add unused bits = 0
         seq[2] = Tlv(seq[2].tag, b'\0' + sig)
         # Re-assemble sequence
         der = Tlv(0x30, b''.join(seq))
 
         return x509.load_der_x509_csr(der, default_backend())
+
+    @property
+    def supports_pin_policies(self):
+        return self.version >= (4, 0, 0)
+
+    @property
+    def supported_touch_policies(self):
+        if self.version < (4, 0, 0):
+            return []  # Touch policy not supported on NEO.
+        elif self.version < (4, 3, 0):
+            return [TOUCH_POLICY.DEFAULT, TOUCH_POLICY.NEVER,
+                    TOUCH_POLICY.ALWAYS]  # Cached policy was added in 4.3
+        else:
+            return [policy for policy in TOUCH_POLICY]
