@@ -38,6 +38,7 @@ from enum import Enum, IntEnum, unique
 from base64 import b32decode
 from binascii import b2a_hex, a2b_hex
 from OpenSSL import crypto
+from .scancodes import KEYBOARD_LAYOUT
 
 
 class BitflagEnum(IntEnum):
@@ -69,22 +70,59 @@ class CAPABILITY(BitflagEnum):
     OPGP = 0x08
     PIV = 0x10
     OATH = 0x20
-    NFC = 0x40
+    FIDO2 = 0x200
 
     @staticmethod
     def dependent_on_ccid():
         return CAPABILITY.OPGP | CAPABILITY.OATH | CAPABILITY.PIV
 
+    def __str__(self):
+        if self == self.U2F:
+            return 'FIDO U2F'
+        elif self == self.FIDO2:
+            return 'FIDO2'
+        elif self == self.OPGP:
+            return 'OpenPGP'
+        else:
+            return self.name
+
+
+@unique
+class FORM_FACTOR(IntEnum):
+    UNKNOWN = 0x00
+    USB_A_KEYCHAIN = 0x01
+    USB_A_NANO = 0x02
+    USB_C_KEYCHAIN = 0x03
+    USB_C_NANO = 0x04
+
+    def __str__(self):
+        if self == self.USB_A_KEYCHAIN:
+            return 'Keychain (USB-A)'
+        elif self == self.USB_A_NANO:
+            return 'Nano (USB-A)'
+        elif self == self.USB_C_KEYCHAIN:
+            return 'Keychain (USB-C)'
+        elif self == self.USB_C_NANO:
+            return 'Nano (USB-C)'
+        elif self == self.UNKNOWN:
+            return 'Unknown.'
+
+    @classmethod
+    def from_code(cls, code):
+        if code and not isinstance(code, int):
+            raise ValueError('Invalid form factor code: {}'.format(code))
+        return cls(code) if code in cls.__members__.values() else cls.UNKNOWN
+
 
 @unique
 class TRANSPORT(BitflagEnum):
     OTP = CAPABILITY.OTP
-    U2F = CAPABILITY.U2F
+    FIDO = CAPABILITY.U2F
     CCID = CAPABILITY.CCID
 
     @staticmethod
     def usb_transports():
-        return TRANSPORT.OTP | TRANSPORT.CCID | TRANSPORT.U2F
+        return TRANSPORT.OTP | TRANSPORT.CCID | TRANSPORT.FIDO
 
 
 class Cve201715361VulnerableError(Exception):
@@ -105,7 +143,7 @@ class Cve201715361VulnerableError(Exception):
 class YUBIKEY(Enum):
     YKS = 'YubiKey Standard'
     NEO = 'YubiKey NEO'
-    SKY = 'FIDO U2F Security Key by Yubico'
+    SKY = 'FIDO U2F Security Key'
     YKP = 'YubiKey Plus'
     YK4 = 'YubiKey 4'
 
@@ -120,19 +158,19 @@ class PID(IntEnum):
     NEO_OTP = 0x0110
     NEO_OTP_CCID = 0x0111
     NEO_CCID = 0x0112
-    NEO_U2F = 0x0113
-    NEO_OTP_U2F = 0x0114
-    NEO_U2F_CCID = 0x0115
-    NEO_OTP_U2F_CCID = 0x0116
-    SKY_U2F = 0x0120
+    NEO_FIDO = 0x0113
+    NEO_OTP_FIDO = 0x0114
+    NEO_FIDO_CCID = 0x0115
+    NEO_OTP_FIDO_CCID = 0x0116
+    SKY_FIDO = 0x0120
     YK4_OTP = 0x0401
-    YK4_U2F = 0x0402
-    YK4_OTP_U2F = 0x0403
+    YK4_FIDO = 0x0402
+    YK4_OTP_FIDO = 0x0403
     YK4_CCID = 0x0404
     YK4_OTP_CCID = 0x0405
-    YK4_U2F_CCID = 0x0406
-    YK4_OTP_U2F_CCID = 0x0407
-    YKP_OTP_U2F = 0x0410
+    YK4_FIDO_CCID = 0x0406
+    YK4_OTP_FIDO_CCID = 0x0407
+    YKP_OTP_FIDO = 0x0410
 
     def get_type(self):
         return YUBIKEY[self.name.split('_', 1)[0]]
@@ -146,10 +184,10 @@ class Mode(object):
         TRANSPORT.OTP,  # 0x00
         TRANSPORT.CCID,  # 0x01
         TRANSPORT.OTP | TRANSPORT.CCID,  # 0x02
-        TRANSPORT.U2F,  # 0x03
-        TRANSPORT.OTP | TRANSPORT.U2F,  # 0x04
-        TRANSPORT.U2F | TRANSPORT.CCID,  # 0x05
-        TRANSPORT.OTP | TRANSPORT.U2F | TRANSPORT.CCID  # 0x06
+        TRANSPORT.FIDO,  # 0x03
+        TRANSPORT.OTP | TRANSPORT.FIDO,  # 0x04
+        TRANSPORT.FIDO | TRANSPORT.CCID,  # 0x05
+        TRANSPORT.OTP | TRANSPORT.FIDO | TRANSPORT.CCID  # 0x06
     ]
 
     def __init__(self, transports):
@@ -277,7 +315,7 @@ _HEX = b'0123456789abcdef'
 _MODHEX = b'cbdefghijklnrtuv'
 _MODHEX_TO_HEX = dict((_MODHEX[i], _HEX[i:i+1]) for i in range(16))
 _HEX_TO_MODHEX = dict((_HEX[i], _MODHEX[i:i+1]) for i in range(16))
-_PW_CHARS = _MODHEX + _MODHEX.upper()
+DEFAULT_PW_CHAR_BLACKLIST = ['\t', '\n', ' ']
 
 
 def ensure_not_cve201715361_vulnerable_firmware_version(f_version):
@@ -299,10 +337,15 @@ def modhex_encode(value):
     return b''.join(_HEX_TO_MODHEX[c] for c in b2a_hex(value)).decode('ascii')
 
 
-def generate_static_pw(length):
+def generate_static_pw(
+    length, keyboard_layout=KEYBOARD_LAYOUT.MODHEX,
+        blacklist=DEFAULT_PW_CHAR_BLACKLIST):
     data = os.urandom(length)
-    return bytes(bytearray(six.indexbytes(_PW_CHARS, d % len(_PW_CHARS))
-                           for d in six.iterbytes(data)))
+    keys = ''.join([
+        k for k in keyboard_layout.value.keys() if k not in blacklist]).encode()
+    return bytes(
+            bytearray(six.indexbytes(
+                keys, d % len(keys)) for d in six.iterbytes(data)))
 
 
 def format_code(code, digits=6, steam=False):
